import copy
import os
import zipfile

import six

from dagster.utils import file_relative_path


def subset_environment_dict(environment_dict, solid_name):
    subset = copy.deepcopy(environment_dict)
    if 'solids' in subset:
        solid_config_keys = list(subset['solids'].keys())
        for key in solid_config_keys:
            if key != solid_name:
                del subset['solids'][key]
    return subset


def build_main_file(
    main_file, mode_name, pipeline_file, solid_name, environment_dict, pipeline_fn_name
):
    with open(file_relative_path(__file__, 'main.py.template'), 'rb') as f:
        main_template_str = six.ensure_str(f.read())

    with open(main_file, 'wb') as f:
        f.write(
            six.ensure_binary(
                main_template_str.format(
                    mode_name=mode_name,
                    pipeline_file=os.path.splitext(os.path.basename(pipeline_file))[0],
                    solid_name=solid_name,
                    environment_dict=subset_environment_dict(environment_dict, solid_name),
                    pipeline_fn_name=pipeline_fn_name,
                )
            )
        )


def build_pyspark_zip(zip_file, path):
    '''Archives the current path into a file named `zip_file`
    '''
    with zipfile.ZipFile(zip_file, 'w', zipfile.ZIP_DEFLATED) as zf:
        for root, _, files in os.walk(path):
            for fname in files:
                abs_fname = os.path.join(root, fname)

                # Skip various artifacts
                if 'pytest' in abs_fname or '__pycache__' in abs_fname or 'pyc' in abs_fname:
                    continue

                zf.write(abs_fname, os.path.relpath(os.path.join(root, fname), path))



# -*- coding: utf-8 -*-
import time

import click
from click.testing import CliRunner
from dagster_aws.cli.term import Spinner, Term


def test_term():
    def term_helper(term_cmd, prefix, exit_code=0):
        @click.command()
        def fn():
            term_cmd('foo bar')

        runner = CliRunner()
        result = runner.invoke(fn)
        assert result.exit_code == exit_code
        assert result.output == prefix + u'foo bar\n'

    expected = [
        (Term.error, Term.ERROR_PREFIX),
        (Term.info, Term.INFO_PREFIX),
        (Term.success, Term.SUCCESS_PREFIX),
        (Term.waiting, Term.WAITING_PREFIX),
        (Term.warning, Term.WARNING_PREFIX),
    ]
    for term_cmd, prefix in expected:
        term_helper(term_cmd, prefix)

    term_helper(Term.fatal, Term.FATAL_PREFIX, exit_code=1)


def test_spinner(capsys):
    with Spinner():
        time.sleep(0.5)

    captured = capsys.readouterr()
    assert captured.out.encode('unicode-escape').startswith(
        b'\\u280b\\x08\\u2819\\x08\\u2839\\x08\\u2838\\x08'
    )



# This file duplicated from the Yelp MRJob project:
#
#   https://github.com/Yelp/mrjob
#
#
# -*- coding: utf-8 -*-
# Copyright 2013 Lyft
# Copyright 2015-2016 Yelp
# Copyright 2017 Yelp and Contributors
# Copyright 2019 Yelp and Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General information about Amazon Web Services, such as region-to-endpoint
mappings. Also includes basic utilities for working with boto3.
"""

import socket
import ssl
from datetime import datetime, timedelta

import botocore
from dateutil.tz import tzutc

from .retry import RetryWrapper

### Utilities ###

# if AWS throttles us, how long to wait (in seconds) before trying again?
_AWS_BACKOFF = 20
_AWS_BACKOFF_MULTIPLIER = 1.5
_AWS_MAX_TRIES = 20  # this takes about a day before we run out of tries


def _client_error_code(ex):
    """Get the error code for the given ClientError"""
    return ex.response.get('Error', {}).get('Code', '')


def _client_error_status(ex):
    """Get the HTTP status for the given ClientError"""
    resp = ex.response
    # sometimes status code is in ResponseMetadata, not Error
    return resp.get('Error', {}).get('HTTPStatusCode') or resp.get('ResponseMetadata', {}).get(
        'HTTPStatusCode'
    )


def _is_retriable_client_error(ex):
    """Is the exception from a boto3 client retriable?"""
    if isinstance(ex, botocore.exceptions.ClientError):
        # these rarely get through in boto3
        code = _client_error_code(ex)
        # "Throttl" catches "Throttled" and "Throttling"
        if any(c in code for c in ('Throttl', 'RequestExpired', 'Timeout')):
            return True
        # spurious 505s thought to be part of an AWS load balancer issue
        return _client_error_status(ex) == 505
    # in Python 2.7, SSLError is a subclass of socket.error, so catch
    # SSLError first
    elif isinstance(ex, ssl.SSLError):
        # catch ssl.SSLError: ('The read operation timed out',). See #1827.
        # also catches 'The write operation timed out'
        return any(isinstance(arg, str) and 'timed out' in arg for arg in ex.args)
    elif isinstance(ex, socket.error):
        return ex.args in ((104, 'Connection reset by peer'), (110, 'Connection timed out'))
    else:
        return False


def _wrap_aws_client(raw_client, min_backoff=None):
    """Wrap a given boto3 Client object so that it can retry when
    throttled."""
    return RetryWrapper(
        raw_client,
        retry_if=_is_retriable_client_error,
        backoff=max(_AWS_BACKOFF, min_backoff or 0),
        multiplier=_AWS_BACKOFF_MULTIPLIER,
        max_tries=_AWS_MAX_TRIES,
        unwrap_methods={'get_paginator'},
    )


def _boto3_now():
    """Get a ``datetime`` that's compatible with :py:mod:`boto3`.
    These are always UTC time, with time zone ``dateutil.tz.tzutc()``.
    """
    if tzutc is None:
        raise ImportError('You must install dateutil to get boto3-compatible datetimes')

    return datetime.now(tzutc())


def strip_microseconds(delta):
    """Return the given :py:class:`datetime.timedelta`, without microseconds.

    Useful for printing :py:class:`datetime.timedelta` objects.
    """
    return timedelta(delta.days, delta.seconds)



import datetime
import json
import time

import pytest
from dagster_aws.cloudwatch import cloudwatch_logger
from dagster_aws.cloudwatch.loggers import millisecond_timestamp

from dagster import ModeDefinition, PipelineDefinition, execute_pipeline, solid

from .conftest import AWS_REGION, TEST_CLOUDWATCH_LOG_GROUP_NAME, TEST_CLOUDWATCH_LOG_STREAM_NAME

TEN_MINUTES_MS = 10 * 60 * 1000  # in milliseconds
NUM_POLL_ATTEMPTS = 5


@solid(name='hello_cloudwatch')
def hello_cloudwatch(context):
    context.log.info('Hello, Cloudwatch!')


def define_hello_cloudwatch_pipeline():
    return PipelineDefinition(
        [hello_cloudwatch],
        name='hello_cloudwatch_pipeline',
        mode_defs=[ModeDefinition(logger_defs={'cloudwatch': cloudwatch_logger})],
    )


def test_cloudwatch_logging_bad_log_group_name():
    with pytest.raises(
        Exception,
        match='Failed to initialize Cloudwatch logger: Could not find log group with name foo',
    ):
        execute_pipeline(
            define_hello_cloudwatch_pipeline(),
            {
                'loggers': {
                    'cloudwatch': {
                        'config': {
                            'log_group_name': 'foo',
                            'log_stream_name': 'bar',
                            'aws_region': 'us-east-1',  # different region
                        }
                    }
                }
            },
        )


def test_cloudwatch_logging_bad_log_stream_name():
    with pytest.raises(
        Exception,
        match='Failed to initialize Cloudwatch logger: Could not find log stream with name bar',
    ):
        execute_pipeline(
            define_hello_cloudwatch_pipeline(),
            {
                'loggers': {
                    'cloudwatch': {
                        'config': {
                            'log_group_name': TEST_CLOUDWATCH_LOG_GROUP_NAME,
                            'log_stream_name': 'bar',
                            'aws_region': AWS_REGION,
                        }
                    }
                }
            },
        )


# TODO: Test bad region


def test_cloudwatch_logging(cloudwatch_client):
    res = execute_pipeline(
        define_hello_cloudwatch_pipeline(),
        {
            'loggers': {
                'cloudwatch': {
                    'config': {
                        'log_group_name': TEST_CLOUDWATCH_LOG_GROUP_NAME,
                        'log_stream_name': TEST_CLOUDWATCH_LOG_STREAM_NAME,
                        'aws_region': AWS_REGION,
                    }
                }
            }
        },
    )

    now = millisecond_timestamp(datetime.datetime.utcnow())

    attempt_num = 0

    found_orig_message = False

    while not found_orig_message and attempt_num < NUM_POLL_ATTEMPTS:
        # Hack: the get_log_events call below won't include events logged in the pipeline execution
        # above if we query too soon after completion.
        time.sleep(1)

        # This is implicitly assuming that we're not running these tests with too much concurrency, etc.
        events = cloudwatch_client.get_log_events(
            startTime=now - TEN_MINUTES_MS,
            logGroupName=TEST_CLOUDWATCH_LOG_GROUP_NAME,
            logStreamName=TEST_CLOUDWATCH_LOG_STREAM_NAME,
            limit=100,
        )['events']

        for parsed_msg in (json.loads(event['message']) for event in events):
            if parsed_msg['dagster_meta']['run_id'] == res.run_id:
                if parsed_msg['dagster_meta']['orig_message'] == 'Hello, Cloudwatch!':
                    found_orig_message = True
                    break

        attempt_num += 1

    assert found_orig_message



from dagster import check
from dagster.core.storage.intermediate_store import IntermediateStore
from dagster.core.storage.type_storage import TypeStoragePluginRegistry

from .object_store import S3ObjectStore


class S3IntermediateStore(IntermediateStore):
    def __init__(
        self,
        s3_bucket,
        run_id,
        s3_session=None,
        type_storage_plugin_registry=None,
        s3_prefix='dagster',
    ):
        check.str_param(s3_bucket, 's3_bucket')
        check.str_param(s3_prefix, 's3_prefix')
        check.str_param(run_id, 'run_id')

        object_store = S3ObjectStore(s3_bucket, s3_session=s3_session)

        def root_for_run_id(r_id):
            return object_store.key_for_paths([s3_prefix, 'storage', r_id])

        super(S3IntermediateStore, self).__init__(
            object_store,
            root_for_run_id=root_for_run_id,
            run_id=run_id,
            type_storage_plugin_registry=check.inst_param(
                type_storage_plugin_registry
                if type_storage_plugin_registry
                else TypeStoragePluginRegistry(types_to_register=[]),
                'type_storage_plugin_registry',
                TypeStoragePluginRegistry,
            ),
        )



import os

import pytest
from dagster_aws.emr import EmrJobRunner, emr_pyspark_resource
from dagster_pyspark import pyspark_resource, pyspark_solid
from moto import mock_emr

from dagster import (
    DagsterInvalidDefinitionError,
    ModeDefinition,
    RunConfig,
    execute_pipeline,
    pipeline,
)
from dagster.seven import mock
from dagster.utils.test import create_test_pipeline_execution_context


@pyspark_solid
def example_solid(context):
    list_p = [('John', 19), ('Jennifer', 29), ('Adam', 35), ('Henry', 50)]
    rdd = context.resources.pyspark.spark_context.parallelize(list_p)
    res = rdd.take(2)
    for name, age in res:
        context.log.info('%s: %d' % (name, age))


@pyspark_solid(name='blah', description='this is a test', config={'foo': str, 'bar': int})
def other_example_solid(context):
    list_p = [('John', 19), ('Jennifer', 29), ('Adam', 35), ('Henry', 50)]
    rdd = context.resources.pyspark.spark_context.parallelize(list_p)
    res = rdd.take(2)
    for name, age in res:
        context.log.info('%s: %d' % (name, age))


@pipeline(
    mode_defs=[
        ModeDefinition('prod', resource_defs={'pyspark': emr_pyspark_resource}),
        ModeDefinition('local', resource_defs={'pyspark': pyspark_resource}),
    ]
)
def example_pipe():
    example_solid()
    other_example_solid()


def test_local():
    result = execute_pipeline(
        example_pipe,
        environment_dict={'solids': {'blah': {'config': {'foo': 'a string', 'bar': 123}}},},
        run_config=RunConfig(mode='local'),
    )
    assert result.success


@mock_emr
@mock.patch('dagster_aws.emr.emr.EmrJobRunner.wait_for_steps_to_complete')
def test_pyspark_emr(mock_wait):
    run_job_flow_args = dict(
        Instances={
            'InstanceCount': 1,
            'KeepJobFlowAliveWhenNoSteps': True,
            'MasterInstanceType': 'c3.medium',
            'Placement': {'AvailabilityZone': 'us-west-1a'},
            'SlaveInstanceType': 'c3.xlarge',
        },
        JobFlowRole='EMR_EC2_DefaultRole',
        LogUri='s3://mybucket/log',
        Name='cluster',
        ServiceRole='EMR_DefaultRole',
        VisibleToAllUsers=True,
    )

    # Doing cluster setup outside of a solid here, because run_job_flow is not yet plumbed through
    # to the pyspark EMR resource.
    job_runner = EmrJobRunner(region='us-west-1')
    context = create_test_pipeline_execution_context()
    cluster_id = job_runner.run_job_flow(context, run_job_flow_args)

    result = execute_pipeline(
        example_pipe,
        environment_dict={
            'solids': {'blah': {'config': {'foo': 'a string', 'bar': 123}}},
            'resources': {
                'pyspark': {
                    'config': {
                        'pipeline_file': __file__,
                        'pipeline_fn_name': 'example_pipe',
                        'cluster_id': cluster_id,
                        'staging_bucket': 'dagster-scratch-80542c2',
                        'region_name': 'us-west-1',
                    }
                }
            },
        },
        run_config=RunConfig(mode='prod'),
    )
    assert result.success
    assert mock_wait.called_once


def test_bad_requirements_txt():
    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        execute_pipeline(
            example_pipe,
            environment_dict={
                'solids': {'blah': {'config': {'foo': 'a string', 'bar': 123}}},
                'resources': {
                    'pyspark': {
                        'config': {
                            'requirements_file_path': 'DOES_NOT_EXIST',
                            'pipeline_file': __file__,
                            'pipeline_fn_name': 'example_pipe',
                            'cluster_id': 'some_cluster_id',
                            'staging_bucket': 'dagster-scratch-80542c2',
                            'region_name': 'us-west-1',
                        }
                    }
                },
            },
            run_config=RunConfig(mode='prod'),
        )
    assert 'The requirements.txt file that was specified does not exist' in str(exc_info.value)


@pytest.mark.skipif(
    'AWS_EMR_TEST_DO_IT_LIVE' not in os.environ,
    reason='This test is slow and requires a live EMR cluster; run only upon explicit request',
)
def test_do_it_live_emr():
    result = execute_pipeline(
        example_pipe,
        environment_dict={
            'solids': {'blah': {'config': {'foo': 'a string', 'bar': 123}}},
            'resources': {
                'pyspark': {
                    'config': {
                        'pipeline_file': __file__,
                        'pipeline_fn_name': 'example_pipe',
                        'cluster_id': os.environ.get('AWS_EMR_JOB_FLOW_ID'),
                        'staging_bucket': 'dagster-scratch-80542c2',
                        'region_name': 'us-west-1',
                        'wait_for_logs': True,
                    }
                }
            },
        },
        run_config=RunConfig(mode='prod'),
    )
    assert result.success



import os

import boto3
import six
from dagster_aws.utils.mrjob.log4j import parse_hadoop_log4j_records
from dagster_pyspark import PySparkResourceDefinition
from dagster_spark.configs_spark import spark_config
from dagster_spark.utils import flatten_dict, format_for_cli

from dagster import Field, check, resource
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.seven import get_system_temp_directory

from .emr import EmrJobRunner
from .utils import build_main_file, build_pyspark_zip

# On EMR, Spark is installed here
EMR_SPARK_HOME = '/usr/lib/spark/'


class EmrPySparkResource(PySparkResourceDefinition):
    '''See the docstring on `emr_pyspark_resource` for instructions on using this resource.
    '''

    def __init__(self, config):
        self.config = config
        self.emr_job_runner = EmrJobRunner(region=self.config['region_name'])
        self.s3_client = boto3.client('s3', region_name=self.config['region_name'])

        # Construct the SparkSession
        super(EmrPySparkResource, self).__init__(self.config.get('spark_conf'))

    def get_compute_fn(self, fn, solid_name):
        '''Construct new compute function for EMR pyspark execution. In the scenario where we are
        running on a Dagster box, we will (1) sync the client code to an S3 staging bucket, and then
        (2) invoke execution via the EMR APIs.

        On EMR, we'll just return the original solid function body to kick off normal pyspark
        execution. Since that will be launched on the EMR master node with YARN, it will
        automatically use the EMR cluster for execution.
        '''

        if self.running_on_emr:
            return fn

        def new_compute_fn(context, *args, **kwargs):  # pylint: disable=unused-argument
            self._sync_code_to_s3(context, solid_name)
            step_defs = self._get_execute_steps(context, solid_name)
            step_ids = self.emr_job_runner.add_job_flow_steps(
                context, self.config['cluster_id'], step_defs
            )
            self.emr_job_runner.wait_for_steps_to_complete(
                context, self.config['cluster_id'], step_ids
            )
            if self.config['wait_for_logs']:
                stdout_log, stderr_log = self.emr_job_runner.retrieve_logs_for_step_id(
                    context, self.config['cluster_id'], step_ids[1]
                )
                # Since stderr is YARN / Hadoop Log4J output, parse and reformat those log lines for
                # Dagster's logging system.
                records = parse_hadoop_log4j_records(stderr_log)
                for record in records:
                    context.log._log(  # pylint: disable=protected-access
                        record.level, record.logger + ': ' + record.message, {}
                    )
                context.log.info(stdout_log)

        return new_compute_fn

    def _sync_code_to_s3(self, context, solid_name):
        '''Synchronize the pyspark code to an S3 staging bucket for use on EMR. Note that
        requirements are installed separately when a requirements.txt is provided.

        For the zip file, consider the following toy example:

        # Folder: my_pyspark_project/
        # a.py
        def foo():
            print(1)

        # b.py
        def bar():
            print(2)

        # main.py
        from a import foo
        from b import bar

        foo()
        bar()

        This will zip up `my_pyspark_project/` as `my_pyspark_project.zip`. Then, when running
        `spark-submit --py-files my_pyspark_project.zip main.py` on EMR this will print 1, 2.

        Note that we also dynamically construct main.py to support targeting execution of a single
        solid on EMR vs. the entire pipeline.
        '''
        run_id = context.run_id
        main_file = os.path.join(get_system_temp_directory(), '%s-main.py' % run_id)
        zip_file = os.path.join(get_system_temp_directory(), '%s-pyspark.zip' % run_id)

        try:
            build_main_file(
                main_file,
                mode_name=context.pipeline_run.mode,
                pipeline_file=self.config['pipeline_file'],
                solid_name=solid_name,
                environment_dict=context.environment_dict,
                pipeline_fn_name=self.config['pipeline_fn_name'],
            )

            build_pyspark_zip(
                zip_file=zip_file,
                path=os.path.dirname(os.path.abspath(self.config['pipeline_file'])),
            )

            self.s3_client.upload_file(
                zip_file, self.config['staging_bucket'], run_id + '/pyspark.zip'
            )
            self.s3_client.upload_file(
                main_file, self.config['staging_bucket'], run_id + '/main.py'
            )

        finally:
            if os.path.exists(main_file):
                os.unlink(main_file)
            if os.path.exists(zip_file):
                os.unlink(zip_file)

    def _get_execute_steps(self, context, solid_name):
        '''From the local Dagster instance, construct EMR steps that will kick off execution on a
        remote EMR cluster.
        '''
        action_on_failure = self.config['action_on_failure']
        staging_bucket = self.config['staging_bucket']

        run_id = context.run_id
        local_root = os.path.dirname(os.path.abspath(self.config['pipeline_file']))

        steps = []

        # Install Python dependencies if a requirements file exists
        requirements_file = self.config.get('requirements_file_path')
        if requirements_file and not os.path.exists(requirements_file):
            raise DagsterInvalidDefinitionError(
                'The requirements.txt file that was specified does not exist'
            )

        if not requirements_file:
            requirements_file = os.path.join(local_root, 'requirements.txt')

        if os.path.exists(requirements_file):
            with open(requirements_file, 'rb') as f:
                python_dependencies = six.ensure_str(f.read()).split('\n')
                steps.append(
                    EmrJobRunner.construct_step_dict_for_command(
                        'Install Dependencies',
                        ['sudo', 'python3', '-m', 'pip', 'install'] + python_dependencies,
                        action_on_failure=action_on_failure,
                    )
                )

        # Execute Solid via spark-submit
        conf = dict(flatten_dict(self.config.get('spark_conf')))
        conf['spark.app.name'] = conf.get('spark.app.name', solid_name)

        check.invariant(
            conf.get('spark.master', 'yarn') == 'yarn',
            desc='spark.master is configured as %s; cannot set Spark master on EMR to anything '
            'other than "yarn"' % conf.get('spark.master'),
        )

        command = (
            [
                EMR_SPARK_HOME + 'bin/spark-submit',
                '--master',
                'yarn',
                '--deploy-mode',
                conf.get('spark.submit.deployMode', 'client'),
            ]
            + format_for_cli(list(flatten_dict(conf)))
            + [
                '--py-files',
                's3://%s/%s/pyspark.zip' % (staging_bucket, run_id),
                's3://%s/%s/main.py' % (staging_bucket, run_id),
            ]
        )

        steps.append(
            EmrJobRunner.construct_step_dict_for_command(
                'Execute Solid %s' % solid_name, command, action_on_failure=action_on_failure
            )
        )
        return steps

    @property
    def running_on_emr(self):
        '''Detects whether we are running on the EMR cluster
        '''
        if os.path.exists('/mnt/var/lib/info/job-flow.json'):
            return True
        return False


@resource(
    {
        'pipeline_file': Field(str, description='Path to the file where the pipeline is defined'),
        'pipeline_fn_name': Field(str),
        'spark_config': spark_config(),
        'cluster_id': Field(str, description='Name of the job flow (cluster) on which to execute'),
        'region_name': Field(str),
        'action_on_failure': Field(str, is_required=False, default_value='CANCEL_AND_WAIT'),
        'staging_bucket': Field(
            str,
            is_required=True,
            description='S3 staging bucket to use for staging the produced main.py and zip file of'
            ' Python code',
        ),
        'requirements_file_path': Field(
            str,
            is_required=False,
            description='Path to a requirements.txt file; the current directory is searched if none'
            ' is specified.',
        ),
        'wait_for_logs': Field(
            bool,
            is_required=False,
            default_value=False,
            description='If set, the system will wait for EMR logs to appear on S3. Note that logs '
            'are copied every 5 minutes, so enabling this will add several minutes to the job '
            'runtime',
        ),
    }
)
def emr_pyspark_resource(init_context):
    '''EMR Pyspark Resource.

    The EMR pyspark resource provides the capability to run pyspark code transparently, without any
    code changes, across both your local machine and EMR. This is as straightforward as the
    following:

    .. code-block:: python

        from dagster import ModeDefinition, pipeline
        from dagster_aws.emr import emr_pyspark_resource
        from dagster_pyspark import pyspark_resource, pyspark_solid

        @pyspark_solid
        def example_solid(context):
            list_p = [('John', 19), ('Jennifer', 29), ('Adam', 35), ('Henry', 50)]
            rdd = context.resources.pyspark.spark_context.parallelize(list_p)
            res = rdd.take(2)
            for name, age in res:
                context.log.info('%s: %d' % (name, age))

        @pipeline(
            mode_defs=[
                ModeDefinition('prod', resource_defs={'pyspark': emr_pyspark_resource}),
                ModeDefinition('local', resource_defs={'pyspark': pyspark_resource}),
            ]
        )
        def example_pipe():
            example_solid()

    When running locally, this will use the standard pyspark resource, which will bring up Spark
    locally and execute your code. In "prod" mode, for each pyspark solid, the EMR resource will:

    1. Sync your code to a zip file on S3, including a code-generated `main.py` to kickstart
       execution;
    2. If a requirements.txt file is present (or specified via resource configuration), Python
       requirements will be installed on EMR prior to job execution;
    3. An EMR job will be constructed for the solid and execution invoked. Job logs will be
    retrieved and logged if configured via `wait_for_logs`.
    '''
    emr_pyspark = EmrPySparkResource(init_context.resource_config)
    try:
        yield emr_pyspark
    finally:
        emr_pyspark.stop()



from .emr import EmrError, EmrJobRunner
from .resources import emr_pyspark_resource
from .types import (
    EMR_CLUSTER_DONE_STATES,
    EMR_CLUSTER_TERMINATED_STATES,
    EmrClusterState,
    EmrStepState,
)



from dagster_aws.emr.utils import subset_environment_dict


def test_subset_environment_dict():
    environment_dict = {
        'solids': {'blah': {'config': {'foo': 'a string', 'bar': 123}}},
        'resources': {
            'pyspark': {
                'config': {
                    'pipeline_file': 'dagster_aws_tests/emr_tests/test_pyspark.py',
                    'pipeline_fn_name': 'pipe',
                    'cluster_id': 'j-272P42200OZ0Q',
                    'staging_bucket': 'dagster-scratch-80542c2',
                    'region_name': 'us-west-1',
                }
            }
        },
    }
    res = subset_environment_dict(environment_dict, 'blah')
    assert res == environment_dict

    res = subset_environment_dict(environment_dict, 'not_here')
    assert res['solids'] == {}



# -*- coding: utf-8 -*-
import os
import subprocess
import sys
import threading
import time
import uuid

import click
import six

from dagster import seven

CURSOR_UP_ONE = '\x1b[1A'
ERASE_LINE = '\x1b[2K'


class Spinner(object):
    '''Spinning CLI prompt, shown while long-running activity is in flight.

    From: https://stackoverflow.com/a/39504463/11295366
    '''

    busy = False
    delay = 0.08
    spinner = ['⠋', '⠙', '⠹', '⠸', '⠼', '⠴', '⠦', '⠧', '⠇', '⠏']

    @staticmethod
    def spinning_cursor():
        while 1:
            for cursor in Spinner.spinner:
                yield cursor

    def __init__(self, delay=None):
        self.spinner_generator = self.spinning_cursor()
        if delay and float(delay):
            self.delay = delay

    def spinner_task(self):
        while self.busy:
            sys.stdout.write(next(self.spinner_generator))
            sys.stdout.flush()
            time.sleep(self.delay)
            sys.stdout.write('\b')
            sys.stdout.flush()

    def __enter__(self):
        self.busy = True
        threading.Thread(target=self.spinner_task).start()

    def __exit__(self, exception, value, tb):
        self.busy = False
        time.sleep(self.delay)
        if exception is not None:
            return False


class Term(object):
    ERROR_PREFIX = u'❌  '
    FATAL_PREFIX = u'💣 '
    INFO_PREFIX = u'ℹ️  '
    SUCCESS_PREFIX = u'✅ '
    WAITING_PREFIX = u'⌛ '
    WARNING_PREFIX = u'⚠️  '

    @staticmethod
    def fatal(msg):
        click.echo(click.style(Term.FATAL_PREFIX + msg, fg='red'), err=True)
        sys.exit(1)

    @staticmethod
    def error(msg):
        click.echo(click.style(Term.ERROR_PREFIX + msg, fg='red'))

    @staticmethod
    def info(msg):
        click.echo(click.style(Term.INFO_PREFIX + msg, fg='blue'))

    @staticmethod
    def success(msg):
        click.echo(click.style(Term.SUCCESS_PREFIX + msg, fg='green'))

    @staticmethod
    def waiting(msg):
        click.echo(click.style(Term.WAITING_PREFIX + msg, fg='yellow'))

    @staticmethod
    def warning(msg):
        click.echo(click.style(Term.WARNING_PREFIX + msg, fg='yellow'))

    @staticmethod
    def rewind():
        sys.stdout.write(CURSOR_UP_ONE)
        sys.stdout.write(ERASE_LINE)


def run_remote_cmd(key_file_path, host, cmd, quiet=False):
    ssh_cmd = 'ssh -o StrictHostKeyChecking=no -i %s ubuntu@%s -t "%s"' % (key_file_path, host, cmd)
    if quiet:
        return subprocess.check_output(ssh_cmd, shell=True)
    else:
        return subprocess.call(ssh_cmd, shell=True)


def rsync_to_remote(key_file_path, local_path, remote_host, remote_path):
    remote_user = 'ubuntu'

    rsync_command = [
        'rsync',
        '-avL',
        '--progress',
        # Exclude a few common paths
        '--exclude',
        '\'.pytest_cache\'',
        '--exclude',
        '\'.git\'',
        '--exclude',
        '\'__pycache__\'',
        '--exclude',
        '\'*.pyc\'',
        '-e',
        '"ssh -o StrictHostKeyChecking=no -i %s"' % key_file_path,
        local_path,
        '%s@%s:%s' % (remote_user, remote_host, remote_path),
    ]
    Term.info('rsyncing local path %s to %s:%s' % (local_path, remote_host, remote_path))
    click.echo('\n' + ' '.join(rsync_command) + '\n')
    subprocess.call(' '.join(rsync_command), shell=True)


def remove_ssh_key(key_file_path):
    # We have to clean up after ourselves to avoid "Too many authentication failures" issue.
    Term.waiting('Removing SSH key from authentication agent...')

    # AWS only gives us the private key contents; ssh-add uses the private key for adding but the
    # public key for removing
    try:
        public_keys = six.ensure_str(subprocess.check_output(['ssh-add', '-L'])).strip().split('\n')
    except subprocess.CalledProcessError:
        Term.rewind()
        Term.info('No identities found, skipping')
        return True

    filtered_public_keys = [key for key in public_keys if key_file_path in key]
    public_key = filtered_public_keys[0] if filtered_public_keys else None

    if public_key:
        tmp_pub_file = os.path.join(
            seven.get_system_temp_directory(), uuid.uuid4().hex + '-tmp-pubkey'
        )

        with open(tmp_pub_file, 'wb') as f:
            f.write(six.ensure_binary(public_key))

        res = subprocess.Popen(
            ['ssh-add', '-d', tmp_pub_file], stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        ).communicate()
        res = six.ensure_str(res[0])

        os.unlink(tmp_pub_file)

        if 'Identity removed' in res:
            Term.rewind()
            Term.success('key deleted successfully')
            return True
        else:
            Term.warning('Could not remove key, error: %s' % res)
            return False
    else:
        Term.rewind()
        Term.info('key not found, skipping')
        return False

    return True



from dagster import Field, String, SystemStorageData, system_storage
from dagster.core.storage.intermediates_manager import IntermediateStoreIntermediatesManager
from dagster.core.storage.system_storage import fs_system_storage, mem_system_storage

from .file_manager import S3FileManager
from .intermediate_store import S3IntermediateStore


@system_storage(
    name='s3',
    is_persistent=True,
    config={
        's3_bucket': Field(String),
        's3_prefix': Field(String, is_required=False, default_value='dagster'),
    },
    required_resource_keys={'s3'},
)
def s3_system_storage(init_context):
    '''Persistent system storage using S3 for storage.
    
    Suitable for intermediates storage for distributed executors, so long as
    each execution node has network connectivity and credentials for S3 and
    the backing bucket.

    Attach this system storage definition, as well as the :py:data:`~dagster_aws.s3_resource` it
    requires, to a :py:class:`~dagster.ModeDefinition` in order to make it available to your
    pipeline:

    .. code-block:: python

        pipeline_def = PipelineDefinition(
            mode_defs=[
                ModeDefinition(
                    resource_defs={'s3': s3_resource, ...},
                    system_storage_defs=default_system_storage_defs + [s3_system_storage, ...],
                    ...
                ), ...
            ], ...
        )

    You may configure this storage as follows:

    .. code-block:: YAML
    
        storage:
          s3:
            config:
              s3_bucket: my-cool-bucket
              s3_prefix: good/prefix-for-files-
    '''
    s3_session = init_context.resources.s3.session
    s3_key = '{prefix}/storage/{run_id}/files'.format(
        prefix=init_context.system_storage_config['s3_prefix'],
        run_id=init_context.pipeline_run.run_id,
    )
    return SystemStorageData(
        file_manager=S3FileManager(
            s3_session=s3_session,
            s3_bucket=init_context.system_storage_config['s3_bucket'],
            s3_base_key=s3_key,
        ),
        intermediates_manager=IntermediateStoreIntermediatesManager(
            S3IntermediateStore(
                s3_session=s3_session,
                s3_bucket=init_context.system_storage_config['s3_bucket'],
                s3_prefix=init_context.system_storage_config['s3_prefix'],
                run_id=init_context.pipeline_run.run_id,
                type_storage_plugin_registry=init_context.type_storage_plugin_registry,
            )
        ),
    )


s3_plus_default_storage_defs = [mem_system_storage, fs_system_storage, s3_system_storage]



from .s3.compute_log_manager import S3ComputeLogManager
from .s3.file_cache import S3FileCache
from .s3.file_manager import S3FileHandle
from .s3.resources import S3Resource, s3_resource
from .s3.s3_fake_resource import S3FakeSession, create_s3_fake_resource
from .s3.solids import S3Coordinate
from .s3.system_storage import s3_plus_default_storage_defs, s3_system_storage
from .version import __version__

__all__ = [
    'create_s3_fake_resource',
    's3_resource',
    's3_system_storage',
    's3_plus_default_storage_defs',
    'S3ComputeLogManager',
    'S3Coordinate',
    'S3FakeSession',
    'S3FileCache',
    'S3FileHandle',
    'S3Resource',
]



from .term import Term, run_remote_cmd


def restart_dagit_service(cfg):
    Term.waiting('Restarting dagit systemd service...')
    retval = run_remote_cmd(cfg.key_file_path, cfg.remote_host, 'sudo systemctl restart dagit')
    Term.rewind()
    if retval == 0:
        Term.rewind()
        Term.success('Service restarted')
    else:
        Term.rewind()
        Term.fatal('Could not restart dagit service')


def start_dagit_service(cfg):
    Term.waiting('Starting dagit systemd service...')
    retval = run_remote_cmd(cfg.key_file_path, cfg.remote_host, 'sudo systemctl start dagit')
    Term.rewind()
    if retval == 0:
        Term.rewind()
        Term.success('Service started')
    else:
        Term.rewind()
        Term.fatal('Could not start dagit service')


def stop_dagit_service(cfg):
    Term.waiting('Stopping dagit systemd service...')
    retval = run_remote_cmd(cfg.key_file_path, cfg.remote_host, 'sudo systemctl stop dagit')
    Term.rewind()
    if retval == 0:
        Term.rewind()
        Term.success('Service stopped')
    else:
        Term.rewind()
        Term.fatal('Could not stop dagit service')



from dagster import Enum, EnumValue

S3ACL = Enum(
    'S3ACL',
    enum_values=[
        EnumValue('private'),
        EnumValue('public-read'),
        EnumValue('public-read-write'),
        EnumValue('authenticated-read'),
        EnumValue('aws-exec-read'),
        EnumValue('bucket-owner-read'),
        EnumValue('bucket-owner-full-control'),
    ],
)



from dagster import Field, resource

from .utils import create_s3_session


class S3Resource(object):
    def __init__(self, s3_session):
        self.session = s3_session

    def download_file(self, bucket, key, filename, **kwargs):
        '''This mirrors the download_file boto3 API:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.download_file
        '''
        return self.session.download_file(bucket, key, filename, **kwargs)

    def download_fileobj(self, bucket, key, fileobj, **kwargs):
        '''This mirrors the download_fileobj boto3 API:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.download_fileobj
        '''
        return self.session.download_fileobj(bucket, key, fileobj, **kwargs)

    def put_object(self, **kwargs):
        '''This mirrors the put_object boto3 API:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Bucket.put_object

        The config schema for this API is applied to the put_object_to_s3_bytes solid vs. at the
        resource level here.
        '''
        return self.session.put_object(**kwargs)

    def upload_fileobj(self, fileobj, bucket, key):
        '''This mirrors the upload_file boto3 API:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.upload_fileobj
        '''
        return self.session.upload_fileobj(fileobj, bucket, key)


@resource(
    {
        'use_unsigned_session': Field(
            bool,
            description='Specifies whether to use an unsigned S3 session',
            is_required=False,
            default_value=False,
        ),
        'region_name': Field(
            str, description='Specifies a custom region for the S3 session', is_required=False
        ),
        'endpoint_url': Field(
            str, description='Specifies a custom endpoint for the S3 session', is_required=False
        ),
    }
)
def s3_resource(context):
    '''Resource that gives solids access to S3.

    The underlying S3 session is created by calling :py:func:`boto3.resource('s3') <boto3:boto3.resource>`.

    Attach this resource definition to a :py:class:`~dagster.ModeDefinition` in order to make it
    available to your solids:

    .. code-block:: python

        pipeline_def = PipelineDefinition(
            mode_defs=[ModeDefinition(resource_defs={'s3': s3_resource, ...}, ...)], ...
        )

    Note that your solids must also declare that they require this resource, or it will not be
    initialized for the execution of their compute functions.

    You may configure this resource as follows:

    .. code-block:: YAML

        resources:
          s3:
            config:
              use_unsigned_session: false
              # Optional[bool]: Specifies whether to use an unsigned S3 session. Default: True
              region_name: "us-west-1"
              # Optional[str]: Specifies a custom region for the S3 session. Default is chosen
              # through the ordinary boto credential chain.
              endpoint_url: "http://localhost"
              # Optional[str]: Specifies a custom endpoint for the S3 session. Default is None.

    '''
    use_unsigned_session = context.resource_config['use_unsigned_session']
    region_name = context.resource_config.get('region_name')
    endpoint_url = context.resource_config.get('endpoint_url')
    return S3Resource(
        create_s3_session(
            signed=not use_unsigned_session, region_name=region_name, endpoint_url=endpoint_url
        )
    )
